@@ -2,6 +2,8 @@
 import logging
 import copy
 
+from random import randint
+
 from snack import SnackScreen
 
 from snackwich.patch import ListboxChoiceWindow, ButtonChoiceWindow, \
@@ -12,22 +14,171 @@ from snackwich.exceptions import GotoPanelException, \
                                  QuitException, \
                                  RedrawException
 
+class _PanelContainer(object):
+    __config  = None
+    __keys    = None
+    __keys_r  = None
+    __ns_keys = None
+
+    def __init__(self, config):
+        """
+        config: List of panel expressions.
+        """
+
+        if config.__class__ != list:
+            message = "Config must be a list."
+
+            logging.error(message)
+            raise TypeError(message)
+
+        self.__config = []
+        self.__keys = []
+        self.__keys_r = {}
+        self.__ns_keys = {}
+
+        logging.info("Loading panels.")
+
+        try:
+            i = 0
+            for expression in config:
+                self.add(expression)
+                i += 1
+        except:
+            logging.exception("Could not load panel (%d)." % (i))
+            raise
+
+    def exists(self, key):
+    
+        return (key in self.__keys_r)
+
+    def get_index_by_key(self, key):
+
+        return self.__keys_r[key]
+
+    def get_copy_by_key(self, key):
+
+        index = self.__keys_r[key]
+        return copy.deepcopy(self.__config[index])
+
+    def remove(self, key, ns=None):
+        if key not in self.__keys_r:
+            raise KeyError("Expression to be removed [%s] is not "
+                           "registered." % (key))
+
+        if ns in self.__ns_keys:
+            if key not in self.__ns_keys[ns]:
+                message = ("Panel [%s] is not indexed under namespace [%s]." % 
+                           (key, ns))
+
+                logging.error(message)
+                raise KeyError(message)
+
+            self.__ns_keys[ns].remove(key)
+            
+            if not self.__ns_keys[ns]:
+                del self.__ns_keys[ns]
+
+        index = self.__keys_r[key]
+        del self.__keys[index]
+        del self.__keys_r[key]
+        del self.__config[index]
+
+    def keys_in_ns(self, ns):
+    
+        return self.__ns_keys[ns][:] if ns in self.__ns_keys else []
+
+    def remove_ns(self, ns):
+
+        logging.info("Removing panels under namespace [%s]." % (ns))
+
+        keys = self.keys_in_ns(ns)
+
+        try:
+            for key in keys:
+                self.remove(key, ns)
+        except:
+            logging.exception("Could not remove all items under namespace "
+                              "[%s]." % (ns))
+            raise
+        
+        return keys
+        
+    def add(self, expression, ns=None, assign_random_key=False):
+
+        try:
+            key = self.__get_current_key(expression)
+        except:
+            logging.exception("Could not render key for expression.")
+            raise
+
+        if assign_random_key:
+            if key != None:
+                message = ("A random key can not be assigned because the key [%s] "
+                           "was already defined." % (key))
+                
+                logging.exception(message)
+                raise KeyError(message)
+
+            while 1:
+                key = ('Random_%d' % randint(11111, 99999))
+                
+                if key not in self.__keys_r:
+                    break
+
+        if not assign_random_key and key in self.__keys_r:
+            raise KeyError("Expression [%s] already registered." % (key))
+
+        new_index = len(self.__keys_r)
+
+        self.__config.append(expression)
+        self.__keys.append(key)
+        self.__keys_r[key] = new_index
+
+        if ns != None:
+            if ns not in self.__ns_keys:
+                self.__ns_keys[ns] = []
+
+            self.__ns_keys[ns].append(key)
+
+        return key
+
+    def __get_current_key(self, expression):
+        """For a given expression, validate the structure and return the key.
+        """
+    
+        if isinstance(expression, tuple):
+            if len(expression) != 2 or \
+               not isinstance(expression[0], str) or \
+               not callable(expression[1]):
+
+                message = "As a tuple, an expression must be " \
+                          "(<string>, <callable>)."
+
+                logging.error(message)
+                raise Exception(message)
+
+            (key, expression_call) = expression
+        elif not isinstance(expression, dict):
+            message = "Expression is not a callable tuple, nor is it a " \
+                      "dictionary."
+
+            logging.error(message)
+            raise Exception(message)
+        else:
+            if '_name' not in expression:
+                message = "'_name' key not found in expression."
+
+                logging.error(message)
+                raise Exception(message)
+
+            key = expression['_name']
+
+        return key
+
 class Snackwich(object):
-    config = None
-
-    # Which panels follows which (dictionary).
-    succession = None
-
-    # The first panel.
-    first_key = None
-
-    # Will describe the key of the panel following the current one.
+    __panels = None
     next_key = None
-
-    # A forward and reverse mapping of expression keys and the expressions' 
-    # indexes within the succession.
-    keys = None
-    keys_r = None
+    results = None
 
     # Some shorthand aliases to make the config nicer. These must be FROM'd 
     # into the current scope, above.
@@ -38,56 +189,14 @@ class Snackwich(object):
                 'message':  MessageWindow,
               }
 
-    def __init__(self, config, forced_succession=None, 
-                 forced_first_key=None, config_overlap=None):
-        """
-        config_filepath: Panel configuration file-path.
-        forced_succession: An specific succession of which panels follow which 
-                           as an alternative to the automatically-calculated 
-                           one. Useful for multiple wizards comprised of a 
-                           subset of the windows having a custom order.
-        forced_first_key: A specific panel to start with instead of simply the 
-                          first one. Again, useful for wizards.
-        config_overlap: Dynamic values that will be written on top of the file-
-                        loaded values.
-        """
-
-        self.succession = forced_succession
-        self.first_key = forced_first_key
-    
-        if config.__class__ != list:
-            message = "Config from file-path [%s] is expected to be a list," \
-                      "not [%s]." % (config_filepath, \
-                                     config.__class__.__name__)
-
-            logging.error(message)
-            raise TypeError(message)
-
-        self.config = config
-
-        logging.info("Mapping panel succession.")
+    def __init__(self, config):
 
         try:
-            (self.keys, self.keys_r) = self.__check_succession()
+            self.__panels = _PanelContainer(config)
         except:
-            logging.exception("Could not calculate panel succession.")
+            logging.exception("Could not build panel contaiment.")
             raise
 
-        if config_overlap:
-            for (key, expression) in config_overlap.iteritems():
-                if key not in self.keys_r:
-                    message = ("Can not overlap panel that does not already "
-                               "exist with name [%s]." % (key))
-
-                    logging.error(message)
-                    raise Exception(message)
-
-                for (expression_key, expression_value) \
-                        in expression.iteritems():
-                
-                    current_expression = self.config[self.keys_r[key]]
-                    current_expression[expression_key] = expression_value
-
     def __process_stanza(self, screen, key, expression, meta_attributes):
         """Process a single expression (form) from the config."""
 
@@ -126,77 +235,6 @@ class Snackwich(object):
 
         return result
 
-    def __get_current_key(self, expression):
-        """For a given expression, validate the structure and return the key.
-        """
-    
-        if isinstance(expression, tuple):
-            if len(expression) != 2 or \
-               not isinstance(expression[0], str) or \
-               not callable(expression[1]):
-
-                message = "As a tuple, an expression must be " \
-                          "(<string>, <callable>)."
-
-                logging.error(message)
-                raise Exception(message)
-
-            (key, expression_call) = expression
-        elif not isinstance(expression, dict):
-            message = "Expression is not a callable tuple, nor is it a " \
-                      "dictionary."
-
-            logging.error(message)
-            raise Exception(message)
-        else:
-            if '_name' not in expression:
-                message = "'_name' key not found in expression."
-
-                logging.error(message)
-                raise Exception(message)
-
-            key = expression['_name']
-
-        return key
-
-    def __check_succession(self):
-        """Run through the configured panel expressions and determine which 
-        leads to which.
-        """
-
-        # Produce the succession of keys in a flat list, first (to avoid 
-        # calculating each key more than once).
-
-        i = 0
-        keys = [ ]
-        keys_r = { }
-        for expression in self.config:
-            try:
-                key = self.__get_current_key(expression)
-            except:
-                logging.exception("Could not render key for expression (%d)." % 
-                                  (i))
-                raise
-
-            keys.append(key)
-            keys_r[key] = i
-            i += 1
-
-        if not self.succession:
-            num_expressions = len(self.config)
-            succession = { }
-            i = 0
-            while i < num_expressions - 1:
-                succession[keys[i]] = keys[i + 1]
-                i += 1
-
-            self.succession = succession
-
-        if not self.first_key:
-            self.first_key = keys[0] if keys else None
-
-        return (keys, keys_r)
-
     def __slice_meta_attributes(self, dict_expression):
         meta_attributes = { }
         to_remove = [ ]
@@ -210,16 +248,76 @@ class Snackwich(object):
 
         return meta_attributes
 
-    def execute(self):
+    def __process_expression(self, expression, results, screen):
+
+        # Allow there to be a function call or lambda that must be 
+        # called to render the actual expression. This call will 
+        # receive the past results as well as the current key. This 
+        # will allow the current form to feed from previous forms.
+        if isinstance(expression, tuple):
+            # The expression is a tuple with the key and a callback.
+        
+            (key, expression_call) = expression
+
+            try:
+                expression = expression_call(self, key, results, 
+                                             screen)
+            except QuitException as e:
+                logging.info("Post-callback for key [%s] has "
+                             "requested emergency exit." % (key))
+                
+                return ('break', key, expression, True)
+            except GotoPanelException as e:
+                # Go to a different panel.
+            
+                new_key = e.key
+                
+                logging.info("We were routed from panel with key [%s] "
+                             "to panel with key [%s]." % (key, 
+                                                          new_key))
+                
+                if not self.__panels.exists(key):
+                    message = "We were told to go to panel with " \
+                              "invalid key [%s] while processing " \
+                              "panel with key [%s]." % (new_key, key)
+                    
+                    logging.error(message)
+                    raise Exception(message)
+                
+                key = new_key
+                return ('continue', key, expression, False)
+            except BreakSuccessionException as e:
+                # Break out of the menu.
+            
+                logging.info("We were told to stop presenting panels "
+                             "while processing panel with key [%s]." % 
+                             (key))
+                return ('break', key, expression, False)
+            except:
+                logging.exception("Callable expression for key [%s] " 
+                                  "threw an exception." % (key))
+                raise
+
+            if not isinstance(expression, dict):
+                message = "Effective expression for key [%s] is not " \
+                          "a dictionary." % (key)
+                
+                logging.exception(message)
+                raise TypeError(message)
+        else:
+            # The expression is a static dictionary.
+
+            key = expression['_name']
+            del expression['_name']
+
+        return (None, key, expression, False)
+
+    def execute(self, first_key):
         """Display the panels. This resembles a state-machine, except that the 
         transitions can be determined on the fly.
         """
 
-        results = { }
-
-        # Make sure we actually had panels.
-        if not self.first_key:
-            return results;
+        self.results = { }
 
         screen = SnackScreen()
 
@@ -227,75 +325,30 @@ class Snackwich(object):
             # Move through the panels in a linked-list fashion, so that we can 
             # reroute if told to.
         
-            key = self.first_key
+            key = first_key
             quit = False
             while 1:
                 logging.info("Processing panel expression with key [%s]." % 
                              (key))
 
-                index = self.keys_r[key]
-                expression = copy.deepcopy(self.config[index])
+                try:
+                    expression = self.__panels.get_copy_by_key(key)
+                except:
+                    logging.exception("Could not retrieve expression [%s]." % 
+                                      (key))
+                    raise
 
-                # Allow there to be a function call or lambda that must be 
-                # called to render the actual expression. This call will 
-                # receive the past results as well as the current key. This 
-                # will allow the current form to feed from previous forms.
-                if isinstance(expression, tuple):
-                    # The expression is a tuple with the key and a callback.
+# Don't reprocess the expression more than once.
+                result = self.__process_expression(expression, self.results, screen)
+                (result_type, key, expression, quit_temp) = result
                 
-                    (key, expression_call) = expression
+                if quit_temp:
+                    quit = True
 
-                    try:
-                        expression = expression_call(self, key, results, 
-                                                     screen)
-                    except QuitException as e:
-                        logging.info("Post-callback for key [%s] has "
-                                     "requested emergency exit." % (key))
-                        
-                        quit = True
-                        break
-                    except GotoPanelException as e:
-                        # Go to a different panel.
-                    
-                        new_key = e.key
-                        
-                        logging.info("We were routed from panel with key [%s] "
-                                     "to panel with key [%s]." % (key, 
-                                                                  new_key))
-                        
-                        if key not in self.keys_r:
-                            message = "We were told to go to panel with " \
-                                      "invalid key [%s] while processing " \
-                                      "panel with key [%s]." % (new_key, key)
-                            
-                            logging.error(message)
-                            raise Exception(message)
-                        
-                        key = new_key
-                        continue
-                    except BreakSuccessionException as e:
-                        # Break out of the menu.
-                    
-                        logging.info("We were told to stop presenting panels "
-                                     "while processing panel with key [%s]." % 
-                                     (key))
-                        break
-                    except:
-                        logging.exception("Callable expression for key [%s] " 
-                                          "threw an exception." % (key))
-                        raise
-
-                    if not isinstance(expression, dict):
-                        message = "Effective expression for key [%s] is not " \
-                                  "a dictionary." % (key)
-                        
-                        logging.exception(message)
-                        raise TypeError(message)
-                else:
-                    # The expression is a static dictionary.
-
-                    key = expression['_name']
-                    del expression['_name']
+                if result_type == 'break':
+                    break
+                elif result_type == 'continue':
+                    continue
 
                 try:
                     meta_attributes = self.__slice_meta_attributes(expression)
@@ -308,20 +361,20 @@ class Snackwich(object):
                 # Determine the panel to succeed this one. We do this early to 
                 # allow callback to adjust this.
 
-                if 'next' in meta_attributes:
-                    next_key = meta_attributes['next']
+                if 'next' not in meta_attributes:
+                    message = "No 'next' panel defined in expression."
                     
-                    if key not in self.keys_r:
-                        logging.error("Key [%s] set as next from panel with "
-                                      "key [%s] does not refer to a valid "
-                                      "panel." % (key))
-                    
-                    self.next_key = next_key
-                else:
-                    if key in self.succession:
-                        self.next_key = self.succession[key]
-                    else:
-                        self.next_key = None
+                    logging.error(message)
+                    raise Exception(message)
+                
+                next_key = meta_attributes['next']
+                
+                if not self.__panels.exists(key):
+                    logging.error("Key [%s] set as next from panel with "
+                                  "key [%s] does not refer to a valid "
+                                  "panel." % (key))
+                
+                self._next_key = next_key
 
                 logging.info("Processing expression with key [%s]." % (key))
 
@@ -357,7 +410,7 @@ class Snackwich(object):
                                          "[%s] to panel with key [%s] while in"
                                          " post callback." % (key, new_key))
                             
-                            if key not in self.keys_r:
+                            if not self.__panels.exists(key):
                                 message = "We were told to go to panel with " \
                                           "invalid key [%s] while in post-" \
                                           "callback for panel with key " \
@@ -377,7 +430,7 @@ class Snackwich(object):
                             # a callback.
                             
                             quit = False
-                            self.next_key = None
+                            self._next_key = None
                             
                             continue;
 
@@ -412,20 +465,20 @@ class Snackwich(object):
 
                     if 'collect_results' in meta_attributes and \
                             meta_attributes['collect_results']:
-                        if key in results:
-                            results[key].append(result)
+                        if key in self.results:
+                            self.results[key].append(result)
                         else:
-                            results[key] = [result]
+                            self.results[key] = [result]
                     else:
-                        results[key] = result
+                        self.results[key] = result
                             
                 except:
                     logging.exception("There was an exception while processing"
                                       " the stanza for key [%s]." % (key))
                     raise
 
-                if self.next_key:
-                    key = self.next_key
+                if self._next_key:
+                    key = self._next_key
                 else:
                     break
         except:
@@ -435,8 +488,33 @@ class Snackwich(object):
         finally:
             screen.finish()
 
-        return None if quit else results
+        return None if quit else self.results
 
     def set_next_panel(self, key):
-        self.next_key = key
+        self._next_key = key
+
+    def get_container(self):
+        return self.__panels
+
+    def get_results_for_ns(self, ns, allow_missing=False):
+        results = {}
+        for key in self.__panels.keys_in_ns(ns):
+            try:
+                results[key] = self.get_results(key)
+            except KeyError:
+                if not allow_missing:
+                    raise
+
+        return results
+                
+    def get_results(self, keys):
+    
+        if keys.__class__ == list:
+            results = {}
+            for key in keys:
+                results[key] = self.results[key]
+
+            return results
+        else:
+            return self.results[keys]
 
